#!/usr/bin/env python

from __future__ import print_function
import sys
from common import *
from Model.model import Model
from Model.dataset import *
import time
from collections import Counter


class EvaluatorMultiRandom:
    def __init__(self, db_type, config, db_config):
        super(EvaluatorMultiRandom, self).__init__()
        self.config = config
        self.db_config = db_config
        self.db_loader, self.dataset = init_dataset(self.config, self.db_config, db_type, False)
        self.dataset_size = len(self.dataset)

        self.gpu_ids = []
        if config.general.gpu_ids != '':
            self.gpu_ids = np.array(self.config.general.gpu_ids.split(' ')).astype(np.int)
        if torch.cuda.is_available() and len(self.gpu_ids) != 0:
            self.device = torch.device('cuda:{0}'.format(self.gpu_ids[0]))
        else:
            self.device = torch.device('cpu')

    def eval(self):
        start = time.time()

        print('evaluating...')

        with torch.no_grad():
            perfects = []
            frames_errors = []

            perfects_b = []
            frames_errors_b = []

            for i, video_data in enumerate(self.db_loader):
                target_prediction = video_data['target_prediction'].to(self.device)
                video_seq_len = video_data['video_seq_len'].to(self.device)

                assert target_prediction.shape[0] == 1

                target_prediction = target_prediction[0].cpu().numpy()

                prediction = np.random.randint(0, video_seq_len[0].cpu().numpy(), len(target_prediction))

                per = (prediction == target_prediction)
                per = per.astype(np.float).sum() / len(per)

                fe = np.abs(
                    prediction - target_prediction)
                fe = fe.astype(np.float).sum() / len(fe)


                perfects.append(per)
                frames_errors.append(fe)

            perfects = np.array(perfects)
            avg_perfects = np.average(perfects)

            frames_errors = np.array(frames_errors)
            avg_frames_error = np.average(frames_errors)

            print('avg_perfects = {}'.format(avg_perfects))
            print('avg_frames_error = {}'.format(avg_frames_error))

        end = time.time()

        print('evaluation took {}'.format(end - start))


def run_once(evaluator):
    evaluator.eval()


def main(argv):
    os.chdir('..')
    config = ConfigParser("Model/config.yaml")
    db_config = ConfigParser("dbs_config.yaml")

    db_name = db_config.general.db_name
    assert db_name == 'lrs2_dataset'

    evaluator = EvaluatorMultiRandom(DBType.Test, config, db_config)
    # multiprocessing.set_start_method('spawn', force=True)
    run_once(evaluator)


if __name__ == '__main__':
    main(sys.argv)
