#!/usr/bin/env python

from __future__ import print_function
import sys
from common import *
from Model.model import Model
from Model.dataset import *
import time
from collections import Counter

######################
# this script evaluates our method with the same scenario as sync net,
# to evaluate syncnet itself, run ./3rdPartiesEvals/syncnet_python/eval.py
######################

class RandomSingleShiftEvaluator:
    def __init__(self, db_type, config, db_config):
        super(RandomSingleShiftEvaluator, self).__init__()
        self.config = config
        self.db_config = db_config
        self.db_loader, self.dataset = init_dataset(self.config, self.db_config, db_type, True)
        self.dataset_size = len(self.dataset)

        self.gpu_ids = []
        if config.general.gpu_ids != '':
            self.gpu_ids = np.array(self.config.general.gpu_ids.split(' ')).astype(np.int)
        if torch.cuda.is_available() and len(self.gpu_ids) != 0:
            self.device = torch.device('cuda:{0}'.format(self.gpu_ids[0]))
        else:
            self.device = torch.device('cpu')

    def eval(self):
        start = time.time()

        print('evaluating...')

        with torch.no_grad():
            perfects = 0
            avg_error = 0
            max_error = 0
            counter = 0

            for i, video_data in enumerate(self.db_loader):
                current_shifts = video_data['shift']

                assert current_shifts.shape[0] == 1

                target_shift = current_shifts[0].cpu().numpy()
                predicted_shift = np.random.randint(0, 25)
                current_error = abs(predicted_shift - target_shift)

                if current_error == 0:
                    perfects += 1

                if current_error > max_error:
                    max_error = current_error

                avg_error += current_error

                counter += 1

                # print('{0}: current error {1}, current avg {2}, shift {3}, acc {4}'.format(counter, current_error,
                #                                                                            float(avg_error) / counter,
                #                                                                            target_shift,
                #                                                                            float(perfects) / counter))

            avg_error = float(avg_error) / counter

            print('====================')
            print('avg_error = {}'.format(avg_error))
            print('====================')
            print('max_error = {}'.format(max_error))
            print('====================')
            print('acc = {}'.format(float(perfects) / counter))
            print('====================')

        end = time.time()

        print('evaluation took {}'.format(end - start))


def run_once(evaluator):
    evaluator.eval()


def main(argv):
    os.chdir('..')
    config = ConfigParser("Model/config.yaml")
    db_config = ConfigParser("dbs_config.yaml")

    db_name = db_config.general.db_name
    assert db_name == 'lrs2_dataset'

    evaluator = RandomSingleShiftEvaluator(DBType.Test, config, db_config)
    # multiprocessing.set_start_method('spawn', force=True)
    run_once(evaluator)


if __name__ == '__main__':
    main(sys.argv)
